# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Helpers."""



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Model exceptions."""


class BaseModelException(Exception):
    """Model-related exceptions inherit from this base class."""

    def __init__(self, msg=None):
        """Initialization."""
        super().__init__()
        self._msg = msg

    @property
    def msg(self):  # pragma: no cover
        """The exception message."""
        return self._msg

    def __str__(self):  # pragma: no cover
        """__str__"""
        return self.msg


class ModelInvalidOperation(BaseModelException):
    """The operation is not possible"""


class ModelConflictException(ModelInvalidOperation):
    """A model instance is conflicting with another model instance."""


class ModelConsistencyException(BaseModelException):
    """The database state is not consistent due to a pending transaction."""



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Request and related classes."""


import json
import asyncio

from aiohttp.web_reqrep import Request as BaseRequest


class Request(BaseRequest):
    """Contains all the information about an incoming HTTP request."""

    def __init__(self, *args, **kwargs):
        """Create a new Request."""
        super().__init__(*args, **kwargs)

    @property
    def args(self):
        """Alternate spelling for `Request.GET`."""
        return self.GET

    @asyncio.coroutine
    def json(self, *, loader=json.loads):
        """Parse the request content as JSON."""
        body = yield from self.text()
        if body:
            return loader(body)
        return None

    @asyncio.coroutine
    def _prepare_hook(self, response):
        #
        # allow early catch of HTTPExceptions before matching a rule
        apps = ()
        if self.match_info:
            apps = self.match_info.apps

        for app in apps:
            yield from app.on_response_prepare.send(self, response)



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Configuration tools."""


import os
import types


class Config(dict):
    """Implements a dictionary with advanced features."""

    def __init__(self, path=None, default=None):
        """initialization.

        Args:
            path: the directory to search for configuration files.
            default: default values in the form of a dictionary.
        """
        dict.__init__(self, default or {})
        self.root_path = path or '.'

    def from_object(self, obj):
        """Load configuration from a python object."""
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)

    def from_pyfile(self, filename, silent=False):
        """Load a configuration file."""
        filename = os.path.join(self.root_path, filename)
        d = types.ModuleType('config')
        d.__file__ = filename
        try:
            with open(filename) as config_file:
                exec(compile(config_file.read(), filename, 'exec'), d.__dict__)
        except IOError as e:
            if not silent:
                e.strerror = 'File not found (%s).' % e.strerror
                raise
        self.from_object(d)

    def from_envvar(self, name, silent=False):
        """Load a configuration file from an environment variable."""
        rv = os.environ.get(name)
        if not rv and not silent:
            err_msg = '%s environment variable is not set.'
            raise RuntimeError(err_msg % name)
        elif rv:
            self.from_pyfile(rv, silent=silent)

    def get_namespace(self, namespace, trim=True):
        """Get a namespace sub-configuration.

        Args:
            namespace: the target namespace.
            trim: trim the namespace from each key.

        Returns: a dictionary
        """
        return {k[(int(trim) * len(namespace)):].lstrip('_'): v for k, v
                in self.items() if k.startswith(namespace)}



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Serializers.

This module provide a set of various serializers along with tools to choose the
best suited one.
"""


from hypr.serializers.accept import choose_media_type
from hypr.serializers.json_serializer import json_serializer


__all__ = ['choose_media_type', 'json_serializer']



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Application factory."""


import asyncio

from aiohttp.web import Application
from aiohttp.log import web_logger
from hypr.request import Request
from hypr.serializers import json_serializer
from hypr.config import Config
from hypr.globals import LocalStorage
from hypr.testing import TestClient
from hypr.router import Router


class Hypr(Application):
    """Hypr runs your application.

    The Hypr object provides a full asynchronous web server based on aiohttp
    and is the corner stone of your application. You will use it to register
    rules, associate them with providers and configure various features.
    """

    config_class = Config
    router_class = Router
    request_class = Request

    serializers = {
        'application/json': json_serializer
    }

    default_config = {
        'DEBUG': False,
        'DEFAULT_MIMETYPE': 'application/json',
        'CRUD_DEFAULT_LIMIT': 10,
        'CRUD_ABSOLUTE_LIMIT': 100,
        'MODELS_SQLITE_DATABASE_URI': 'file:/dev/null?mode=memory&cache=shared'
    }

    def __init__(self, *, logger=web_logger, loop=None,
                 router=None,
                 middlewares=(), debug=...):
        """Initialize the application."""
        if router is None:
            router = self.router_class(self)

        super().__init__(logger=logger, loop=loop, router=router,
                         middlewares=middlewares, debug=debug)

        LocalStorage.bind(self)     # initialize the context storage

        self.config = self.config_class(default=self.default_config)
        self.router.set_serializers(self.serializers)

    @property
    def available_mimetypes(self):
        """Return available mimetypes."""
        # TODO: improve performances by caching the result
        default_mimetype = self.config.get('DEFAULT_MIMETYPE')

        rv = [default_mimetype]
        rv.extend(self.serializers.keys())
        return rv

    def register_on_request_teardown(self, func, *args, **kwargs):
        """Register callbacks to execute on the request teardown."""
        callbacks = LocalStorage().get('_request_teardown_callbacks', None)
        if callbacks is None:
            callbacks = []
            LocalStorage().set('_request_teardown_callbacks', callbacks)
        callbacks.insert(0, (func, args, kwargs))

    def request_teardown(self):
        """Execute callbacks on the request teardown."""
        callbacks = LocalStorage().get('_request_teardown_callbacks', [])
        for (cb, args, kwargs) in callbacks:
            # TODO: handle exceptions
            res = cb(*args, **kwargs)
            if (asyncio.iscoroutine(res) or isinstance(res, asyncio.Future)):
                yield from res

    def add_provider(self, provider, *urls, methods=None, endpoint=None):
        """Register a provider.

        Args:
            provider: a Provider class.
            *urls: one or more path to access the provider.
            methods: a tuple containing the allowed HTTP methods.
                If methods is none, the allowed methods are determined from
                the provider implementation.
            endpoint: the name of the endpoint. If endpoint is none, the name
                is the provider's name.
        """
        self.router.add_provider(provider, *urls, methods=methods,
                                 endpoint=endpoint, propagation=True)

    def provide(self, *urls):
        """A decorator used to register a provider with URL rules.

        This does the same thing as add_provider() but is intended for
        decorator usage:

            @app.provide('/')
            class SomeProvider(Provider):
                ...

        See the documentation for add_provider() for more informations.
        """
        def decorator(provider):
            self.add_provider(provider, *urls)
            return provider
        return decorator

    def test_client(self):
        """Get a test client for this application.

        Being able to test your application is a mandatory requirement if you
        want to provide your users with trustful software. The test_client()
        aims to facilitate the testability of it by providing a full
        implementation of the aiohttp client API.

        The following example illustrates how to use it in a test:

            def test_get_response(self, app):

                with app.test_client() as client:
                    rv = client.get('/test')
                    assert rv.status == 200

        Where `app` is an instance of the application under test.
        """
        return TestClient(self)

    def _make_request(self, message, payload, protocol):
        return self.request_class(
            message, payload,
            protocol.transport, protocol.reader, protocol.writer,
            protocol.time_service, protocol._request_handler,
            secure_proxy_ssl_header=self._secure_proxy_ssl_header
        )

    def make_handler(self, *, secure_proxy_ssl_header=None, **kwargs):
        # Apply declarative propagations before the server startup.
        self.router._propagate()
        return super().make_handler(
            secure_proxy_ssl_header=secure_proxy_ssl_header, **kwargs)

    def run(self, host='127.0.0.1', port=5555, **options):  # pragma: no cover
        """
        Run the application as a web server.

        The application will run as a web server indefinitely until being
        interrupted with a Ctrl-C or a signal.

        Args:
            host: the hostname or IP adress of the application.
            port: the port to listen to.
        """
        loop = self.loop
        handler = self.make_handler(**options)

        task = loop.create_server(handler, host, port)
        server = loop.run_until_complete(task)

        try:
            loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            loop.run_until_complete(handler.finish_connections(1.0))
            server.close()
            loop.run_until_complete(server.wait_closed())
            loop.run_until_complete(self.finish())
        loop.close()



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Web exceptions."""


from aiohttp.web_exceptions import *


error_codes = {

    # 3xx redirection

    300: HTTPMultipleChoices,
    301: HTTPMovedPermanently,
    302: HTTPFound,
    303: HTTPSeeOther,
    304: HTTPNotModified,
    305: HTTPUseProxy,
    307: HTTPTemporaryRedirect,

    # 4xx client error

    400: HTTPBadRequest,
    401: HTTPUnauthorized,
    402: HTTPPaymentRequired,
    403: HTTPForbidden,
    404: HTTPNotFound,
    405: HTTPMethodNotAllowed,
    406: HTTPNotAcceptable,
    407: HTTPProxyAuthenticationRequired,
    408: HTTPRequestTimeout,
    409: HTTPConflict,
    410: HTTPGone,
    411: HTTPLengthRequired,
    412: HTTPPreconditionFailed,
    413: HTTPRequestEntityTooLarge,
    414: HTTPRequestURITooLong,
    415: HTTPUnsupportedMediaType,
    416: HTTPRequestRangeNotSatisfiable,
    417: HTTPExpectationFailed,

    # 5xx server error

    500: HTTPInternalServerError,
    501: HTTPNotImplemented,
    502: HTTPBadGateway,
    503: HTTPServiceUnavailable,
    504: HTTPGatewayTimeout,
    505: HTTPVersionNotSupported,

}


def abort(code, *args, **kwargs):
    """Abort the request processing and return an error to the client."""
    if code in range(400, 418) or code in range(500, 506):
        raise error_codes[code](*args, **kwargs)
    raise HTTPInternalServerError()


def redirect(location, code=302):
    """Stop request processing and redirect the client."""
    if code in range(300, 308):
        raise error_codes[code](location)

    raise HTTPInternalServerError()



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""URL rule converters."""

import re
import uuid


class BaseConverter:
    """Base class for all converters."""

    regex = '[^/]+'
    weight = 100

    def to_python(self, value):
        """Return the value of the segment."""
        return value


class UnicodeConverter(BaseConverter):
    """Unicode converter.

    This converter is the default converter and accepts any string but only one
    path segment.  Thus the string can not include a slash. This is the default
    validator.

    Example:

        Rule('/pages/<page>'),
        Rule('/<string(length=2):lang_code>')

    Args:
        minlength (int): the minimum length of the string.  Must be greater
            or equal 1. (default: 1)
        maxlength (int): the maximum length of the string. (default: None)
        length (int): the exact length of the string. (default: None)
    """

    def __init__(self, minlength=1, maxlength=None, length=None):
        """Init the converter."""
        super().__init__()
        if length is not None:
            length = '{%d}' % int(length)
        else:
            if maxlength is None:
                maxlength = ''
            else:
                maxlength = int(maxlength)
            length = '{%s,%s}' % (int(minlength), maxlength)

        self.regex = '[^/]' + length


class AnyConverter(BaseConverter):
    """Any converter.

    Matches one of the items provided.  Items can either be Python identifiers
    or strings.

        Rule('/<any(about, help, imprint, class, "foo,bar"):page_name>')

    Args:
        items: this function accepts the possible items as positional
            arguments.
    """

    weight = 75

    def __init__(self, *items):
        """Init the converter."""
        super().__init__()
        self.regex = '(?:%s)' % '|'.join([re.escape(x) for x in items])


class PathConverter(BaseConverter):
    """Path converter.

    Like the default :class:`UnicodeConverter`, but it also matches slashes.
    This is useful for wikis and similar applications.

        Rule('/<path:wikipage>')
        Rule('/<path:wikipage>/edit')
    """

    regex = '[^/].*?'
    weight = 200


class NumberConverter(BaseConverter):
    """Baseclass for IntegerConverter and FloatConverter."""

    weight = 50

    def __init__(self, fixed_digits=0, min=None, max=None):
        """Init converter."""
        super().__init__()
        self.fixed_digits = fixed_digits
        self.min = min
        self.max = max

    def to_python(self, value):
        """Convert matched value to the rightful python type."""
        if (self.fixed_digits and len(value) != self.fixed_digits):
            raise ValueError()
        value = self.num_convert(value)
        if (self.min is not None and value < self.min) or \
           (self.max is not None and value > self.max):
            raise ValueError()
        return value


class IntegerConverter(NumberConverter):
    """Integer converter.

    This converter only accepts integer values.

        Rule('/page/<int:page>')

    Args:
        fixed_digits: the number of fixed digits in the URL.  If you set this
            to 4 for example, the application will only match if the url looks
            like /0001. The default is variable length.
        min: the minimal value.
        max: the maximal value.
    """

    regex = r'-?\d+'
    num_convert = int


class FloatConverter(NumberConverter):
    """Float converter.

    This converter only accepts floating point values.

        Rule('/probability/<float:probability>')

    Args:
        min: the minimal value.
        max: the maximal value.
    """

    regex = r'-?(\d+.)?\d+(e[-+]\d+)?'
    num_convert = float

    def __init__(self, min=None, max=None):
        """Init converter."""
        super().__init__(0, min, max)


class UUIDConverter(BaseConverter):
    """UUID converter.

    This converter accepts valid UUIDs.

        Rule('/object/<uuid:identifier>')
    """

    weight = 80

    regex = r'[A-Fa-f0-9]{8}-[A-Fa-f0-9]{4}-' \
            r'[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{12}'

    def to_python(self, value):
        """Convert matched value to the rightful python type."""
        return uuid.UUID(value)


DEFAULT_CONVERTERS = {
    'default':          UnicodeConverter,
    'string':           UnicodeConverter,
    'any':              AnyConverter,
    'path':             PathConverter,
    'int':              IntegerConverter,
    'float':            FloatConverter,
    'uuid':             UUIDConverter,
}



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Available models."""


import sys
import pkg_resources as pkg_resources
from hypr.models.sqlite import SQLiteModel

available_models = ['SQLiteModel']

for ep in pkg_resources.iter_entry_points('hypr.models'):
    model = ep.load()
    setattr(sys.modules[__name__], model.__name__, model)
    available_models.append(model.__name__)

__all__ = tuple(available_models)



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""JSON serializer."""
import json
from datetime import tzinfo, timedelta


class _simple_utc(tzinfo):  # pragma: no cover

    def tzname(self):
        return "UTC"

    def utcoffset(self, dt):
        return timedelta(0)


def encoder(obj):
    """Json encoder."""
    # handle date[time] objects
    if hasattr(obj, 'strftime'):
        if hasattr(obj, 'hour'):
            return obj.replace(tzinfo=_simple_utc()).isoformat()
        else:
            return '%sT00:00:00+00:00' % obj.isoformat()

    # handle uuid objects
    if hasattr(obj, 'hex'):
        return str(obj)

    # handle serializable objects
    if hasattr(obj, '__serialize__'):
        return obj.__serialize__()

    raise TypeError('%s is not JSON serializable.' % repr(obj))


def json_serializer(obj, request=None, default=None):
    """Json serializer."""
    pretty_print = False
    if request is not None:
        pretty_print = request.args.get('_pprint', '').lower() in ('true', '1')

    indent = None
    if pretty_print:    # pragma: no cover
        indent = 4

    if default is None:
        default = encoder

    return json.dumps(obj, default=default, sort_keys=True, indent=indent)



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""SQLite3 Model.

The SQLite3 model is a basic model for data persistence with nearly no features
nor optimizations and based on the sqlite3 module included in the standard
library of Python.

Do not use this model in production!
"""


import pickle
import sqlite3
from hypr.helpers.mini_dsl import normalize_query
from hypr.models.exc import ModelInvalidOperation, ModelConflictException
from hypr.models.base import BaseModel
from hypr.globals import LocalStorage


locals = LocalStorage()


class _ModelType(type):
    def __new__(mcs, name, bases, d):
        cls = type.__new__(mcs, name, bases, d)

        # keep a track of models with a tablename for later initialization
        if cls.__tablename__ is not None:
            SQLiteModel._known_tables.append(cls.__tablename__)
        return cls


class SQLiteModel(BaseModel, metaclass=_ModelType):
    """A model based on SQLite3.

    Supported features:
    - transactional model
    - create, read, update, delete
    - filtering, limiting and sorting of the returned instances

    The SQLiteModel does not support property mapping, querying and relational
    features of a SQL database and is not optimized. Being CPU-bound and fully
    synchronous, a request on a model with a few instances of it stored could
    result on long processing times.
    """

    _db = None
    _root_session = None
    _known_tables = []

    __key__ = 'id',
    __tablename__ = None

    @staticmethod
    def _deserialize(id, blob):
        # deserialize the blob stored in the database
        rv = pickle.loads(blob)
        setattr(rv, '_id', id)
        return rv

    @classmethod
    def _autobind(cls):
        # automatically try to bind the model to the database declared in Hypr
        # configuration "MODELS_SQLITE_DATABASE_URI"
        if SQLiteModel._db is None and locals._app is not None:
            url = locals._app.config.get('MODELS_SQLITE_DATABASE_URI', None)
            if url is not None:
                SQLiteModel.bind(url)

    @classmethod
    def _session(cls):
        # return a session.
        cls._autobind()
        if SQLiteModel._db is None:
            raise RuntimeError('SQLiteModel is not bound to a database.')

        try:
            session = locals.get('_sqlite_localsession', None)
        except RuntimeError:
            if SQLiteModel._root_session is None:
                SQLiteModel._root_session = sqlite3.connect(
                    SQLiteModel._db, uri=True, isolation_level='DEFERRED')
            session = SQLiteModel._root_session
        else:
            if session is None:
                session = sqlite3.connect(
                    SQLiteModel._db, uri=True, isolation_level='DEFERRED')
                locals.set('_sqlite_localsession', session)
            current_app = locals.app()
            current_app.register_on_request_teardown(session.close)

        return session

    @classmethod
    def bind(cls, url):
        """Bind the models to a database."""
        SQLiteModel._db = url

    @classmethod
    def create_all(cls):
        """Create all required tables."""
        session = cls._session()
        q = 'CREATE TABLE IF NOT EXISTS %s (id INTEGER PRIMARY KEY, obj BLOB)'

        for tablename in cls._known_tables:
            session.execute(q % tablename)
        session.commit()

    @property
    def id(self):
        """The resource id."""
        return getattr(self, '_id', None)

    @classmethod
    def get(cls, _offset=0, _limit=-1, _sort=None, **kwargs):
        """Get a list of instances."""
        if _sort is None:
            _sort = ()
        elif not isinstance(_sort, tuple):
            _sort = _sort,

        # Ensure the filters (kwargs) are in the HMQ format.
        for k, v in kwargs.items():
            kwargs[k] = normalize_query(v, key=k)

        # get all instances
        c = cls._session().cursor()
        query = 'SELECT id, obj FROM %s' % cls.__tablename__
        instances = (cls._deserialize(*i) for i in c.execute(query))

        # filter by using the default _match() method. Complexity is O(n) !
        instances = [i for i in instances if i._match(**kwargs)]

        # apply sorting
        # the sort arguments order is reversed to reproduce the behaviour of
        # SQL (ORDER BY A, B === sort by A then by B for each value of A).
        for string in reversed(_sort):
            reverse = string[0] == '-'
            key = string[reverse:]
            instances.sort(key=lambda i: getattr(i, key), reverse=reverse)

        # pagination
        if _limit > 0:
            return instances[_offset:_offset + _limit]

        return instances

    @classmethod
    def one(cls, *args):
        """Get one resource."""
        rv = None

        if cls.__key__ == ('id',):  # tiny optimization to limit damages.
            msg = 'The keys does not ensure the uniqueness of an instance.'
            query = 'SELECT id, obj FROM %s WHERE id = ?' % cls.__tablename__

            c = cls._session().cursor()
            for row in c.execute(query, args):
                if rv is not None:  # pragma: no cover
                    # this shouldn't happen if id is the table's primary key
                    raise NotImplementedError(msg)
                rv = cls._deserialize(*row)
        else:
            rv = super().one(*args)  # use default unoptimized version.
        return rv

    def save(self, commit=True):
        """Make the instance persistent in the data store.

        Args:
            commit: Perform a commit at the same time.
        """
        obj = pickle.dumps(self)
        c = self._session().cursor()

        if self.id is None:
            query = 'INSERT INTO %s (obj) VALUES (?)'
            c.execute(query % self.__tablename__, (obj,))
        else:
            query = 'UPDATE %s SET obj = ? WHERE id = ?'
            c.execute(query % self.__tablename__, (obj, self.id))
            if c.rowcount == 0:
                raise ModelConflictException('instance not found')

        self._id = c.lastrowid or self.id
        if commit:
            self.commit()

        return self

    def delete(self, commit=True):
        """Delete the instance from the data store.

        Args:
            commit: Perform a commit at the same time.
        """
        if self.id is None:
            raise ModelInvalidOperation('instance not persistent')

        query = 'DELETE FROM %s WHERE id = ?'

        c = self._session().cursor()
        c.execute(query % self.__tablename__, (self.id,))
        if c.rowcount == 0:
            raise ModelConflictException('instance not found')

        if commit:
            self.commit()

        return self

    @classmethod
    def commit(cls):
        """Commit changes."""
        cls._session().commit()

    @classmethod
    def rollback(cls):
        """Rollback changes."""
        cls._session().rollback()



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Testing tools."""

from aiohttp import request
import asyncio
import socket


def get_unused_port():
    """Return a unused port on localhost."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('127.0.0.1', 0))
    port = s.getsockname()[1]
    s.close()
    return port


class TestClient:
    """Hypr application test client.

    The test client allow a user to perform HTTP requests to a Hypr application
    by providing the full client interface of aiohttp. It works as a context
    manager to automatically clean-up everything whean leaving it.
    """

    def __init__(self, app):
        """Init a test client.

        Args:
            app: a Hypr application.
        """
        self.app = app
        self.srv = None
        self.port = None
        self.handler = self.app.make_handler()

    @asyncio.coroutine
    def create_server(self):
        """Create a server."""
        self.port = get_unused_port()
        self.srv = yield from self.app.loop.create_server(
            self.handler, '127.0.0.1', self.port)

    def __enter__(self):
        """Entering the context manager."""
        if self.srv is None:
            self.app.loop.run_until_complete(self.create_server())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Leaving the context manager."""
        if self.srv is not None:
            self.srv.close()
            self.app.loop.run_until_complete(self.handler.finish_connections())
            self.srv = None

    def request(self, method, path, **kwargs):
        """Perform an HTTP request.

        Args:
            method (str): HTTP method.
            url (str): Requested URL.
            kwargs: see ``aiohttp.request`` available parameters.

        Returns:
            Return a ``client response`` object.
        """
        kwargs['loop'] = self.app.loop

        @asyncio.coroutine
        def do_request():

            if self.srv is None:
                yield from self.create_server()  # pragma: no cover

            assert path.startswith('/'), '`path` should start with a \'/\'.'

            kwargs['url'] = 'http://127.0.0.1:{}'.format(self.port) + path
            rv = yield from request(method, **kwargs)
            rv.text = yield from rv.text()
            return rv

        return self.app.loop.run_until_complete(do_request())

    def get(self, path, **kwargs):
        """Perform a HTTP GET request."""
        return self.request('GET', path, **kwargs)

    def head(self, path, **kwargs):
        """Perform a HTTP HEAD request."""
        return self.request('HEAD', path, **kwargs)

    def options(self, path, **kwargs):
        """Perform a HTTP OPTIONS request."""
        return self.request('OPTIONS', path, **kwargs)

    def post(self, path, **kwargs):
        """Perform a HTTP POST request."""
        return self.request('POST', path, **kwargs)

    def put(self, path, **kwargs):
        """Perform a HTTP PUT request."""
        return self.request('PUT', path, **kwargs)

    def patch(self, path, **kwargs):
        """Perform a HTTP PACTH request."""
        return self.request('PATCH', path, **kwargs)

    def delete(self, path, **kwargs):
        """Perform a HTTP DELETE request."""
        return self.request('DELETE', path, **kwargs)



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Implements the base model."""


import inspect
from collections import defaultdict


class BaseModel:
    """Base class of all models."""

    @classmethod
    def _key(cls):
        """Get the key of a model.

        The key is a set of properties used to determine the unique identifier
        of a resource.

        Returns:
            a tuple.
        """
        if not hasattr(cls, '__key__'):  # pragma: no cover
            raise NotImplementedError()

        rv = cls.__key__
        if not isinstance(rv, tuple):   # ensure the returned value is a tuple
            rv = rv,

        return rv

    @property
    def _uid(self):
        """Get the unique identifier of a resource.

        The unique identifier (uid) of a resource is a set of values unique to
        each instance of a model.

        Returns:
            a tuple.
        """
        return tuple(getattr(self, k) for k in self._key())

    def _match(self, **kwargs):
        # utterly complex method indicating if an object match a set of filters
        # in HMQ format.

        disjunctions = defaultdict(list)

        # reorder the query by disjunction groups and check for unknown filters
        for key, values in kwargs.items():
            for positive, value, group in values:
                disjunctions[group].append((positive, key, value))

        for group in disjunctions.values():     # for each disjunction group
            for positive, key, value in group:  # for each proposition

                attr = getattr(self, key, None)
                start = getattr(value, 'start', None)
                stop = getattr(value, 'stop', None)

                if hasattr(value, 'match'):
                    if positive and value.match(attr):
                        break
                    if not positive and value.match(attr) is None:
                        break
                elif positive:
                    if start is not None and attr >= start:
                        if stop is not None and attr < stop:
                            break
                        if stop is None:
                            break
                    if start is None and stop is not None and attr < stop:
                        break
                    if attr == value:
                        break
                else:
                    if start is not None and attr < start:
                        break
                    elif stop is not None and attr >= stop:
                        break
                    elif start is None and stop is None and attr != value:
                        break

            else:
                break

        else:
            return True

        return False

    @classmethod
    def get(cls, _offset, _limit, _sort, **kwargs):  # pragma: no cover
        """Get a list of resources."""
        raise NotImplementedError()

    @classmethod
    def count(cls, **kwargs):
        """Get the number of resources."""
        # This is a basic and not optimized method based on get()
        return len(cls.get(**kwargs))

    @classmethod
    def one(cls, *args):
        """Get one resource."""
        # This is a basic and not optimized method based on get()
        col = cls._key()
        if len(args) != len(col):
            raise ValueError('Invalid key')
        rv = cls.get(**{k: v for k, v in zip(col, args)})
        if len(rv) == 1:
            return rv[0]
        else:
            msg = 'The keys does not ensure the uniqueness of an instance.'
            raise NotImplementedError(msg)

    def save(self, commit=True):  # pragma: no cover
        """Make the resource persistent in the data store.

        Args:
            commit: commit inline.
        """
        raise NotImplementedError()

    def delete(self):  # pragma: no cover
        """Delete the resource from the data store.

        Args:
            commit: commit inline.
        """
        raise NotImplementedError()

    @classmethod
    def commit(cls):  # pragma: no cover
        """Commit changes."""
        raise NotImplementedError()

    @classmethod
    def rollback(cls):  # pragma: no cover
        """Rollback changes.

        The rollback is usable on the pending or failed commits.
        """
        pass

    def __serialize__(self, context=None):
        """Serialize the instance.

        Args:
            context: change the behaviour of the serialization process.
        """
        return {k: v for k, v in inspect.getmembers(self)
                if not k.startswith('_') and not inspect.isroutine(v)}



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Implements the base provider."""


import asyncio
import inspect

from hypr.router import Router
from collections import defaultdict


# frozenset of the allowed HTTP methods
HTTP_METHODS = frozenset(['get', 'post', 'head', 'options', 'delete', 'put',
                          'trace', 'patch'])


def _sort(checkpoints):
    # checkpoints is a dictionary where key is the function name of the CP
    # and the value is a tuple (priority, methods). The CP are sorted by
    # proprity, then a list of tuples name, methods is returned.
    checkpoints = sorted(checkpoints.items(), key=lambda x: x[1][0])
    return tuple((name, methods) for name, (_, methods) in checkpoints)


class _ProviderType(type):
    # Perform some class initialization like detecting available HTTP methods
    # and setting up the internal router.

    def __new__(mcs, name, bases, d):

        cls = type.__new__(mcs, name, bases, d)

        methods = set()
        checkpoints = defaultdict(dict)

        # get all the name of propagation rules in the current provider
        rules = ()
        if cls.propagation is not None:
            rules = tuple(name for name in cls.propagation.keys()
                          if not name.startswith('__pps_'))

        for name, func in inspect.getmembers(cls, inspect.isfunction):

            if name in HTTP_METHODS:
                methods.add(name.upper())

            # get the scope and settings of checkpoints
            scope, *settings = getattr(func, '__is_checkpoint', ((),))
            for element in scope:
                # the checkpoint is executed when the rule is applied
                if not isinstance(element, int):
                    checkpoints['__src_%s' % element][name] = settings
                    checkpoints['__lnk_%s' % element][name] = settings
                    continue

                if element & 0b1010 == 0b1010:
                    checkpoints['__sngl'][name] = settings

                if element & 0b0101 == 0b0101:
                    for rule in rules:
                        checkpoints['__lnk_%s' % rule][name] = settings

                if element & 0b1001 == 0b1001:
                    for rule in rules:
                        checkpoints['__src_%s' % rule][name] = settings

                if element & 0b0110 == 0b0110:
                    checkpoints['__dest'][name] = settings

        # flatten and sort checkpoints = cp[scope] = [(func, methods), ...]
        cls._checkpoints = {s: _sort(cp) for s, cp in checkpoints.items()}

        cls.methods = cls.methods or tuple(methods)
        cls._router = Router(None)

        return cls


class Provider(metaclass=_ProviderType):
    """Base class of all providers.

    The provider has the responsibility to dispatch a request to the rightful
    method in order to get the response. The methods able to provide a response
    are named after the HTTP methods (see HTTP_METHODS).

    Propagation
    -----------

    The provider is able to delegate the request processing to another provider
    with the help of a propagation mechanism. The propagation allows to
    register a path to another provider relative to this one.

    Three methods are available to declare a propagation.

    * Declarative method :

    The declarative method is the preferred one as it allows to perform static
    analysis of it. To declare a propagation, each provider class possesses a
    `propagation` attribute where it is possible to list rules. The attribute
    is a dict, a rule is a key-value pair where the key is the endpoint of the
    propagation rule and the value is a tuple with the target provider as the
    first element of it, and URL rules are the subsequent elements.

        class SourceProvider(Provider):

            propagation = {
                'rule0' : (TargetProvider, '/target'),
                'rule1' : OtherProvider
            }

            ...

    If no URL rules are specified, the propagation rule uses the URLs rules
    associated at the app-level with registered target provider. The tuple
    notation is not mandatory for this latter use case. All the propagation
    rules set with the declarative method are resolved at the application
    startup.

    * Imperative method :

    The imperative method allows to declare a propagation rule at the runtime.

        SourceProvider.add_propagation(TargetProvider, '/target')

    This method is potentially harmful and must be used with caution.

    * Decorative method :

    The decorative method is a class decorator for the provider to declare
    a propagation. Optional arguments allow to specify various settings. The
    mechanism of the decorator is based on the declarative method and does not
    require specific usage cautions.

        @SourceProvider.propagate
        class TargetProvider(Provider):

            ...

    Both of declarative and imperative methods support the use of app-level
    registered endpoints instead of providers as target for the propagation.

    Security
    --------

    The security of the provider is achieved with the help of methods
    considered as security checkpoints. To mark a method as a security
    checkpoint the decorator @checkpoint must be used:

        @checkpoint
        def security_checkpoint(self):
            ...

    The decorated method is called by the provider before continuing with
    further tasks and is responsible, with the help of abort(), to stop the
    request processing if the wanted conditions are not met. The decorated
    method accepts arguments the same way as any other HTTP method.

    Each security checkpoint possesses attributes to set when to call it (see
    checkpoint() documentation for mor informations).
    """

    # The checkpoint's scope is encoded on 4 bits and is set with these flags :

    ALWAYS = 0b1111         # always run the checkpoint (CP)

    IS_DEST = 0b1110        # run the CP if the provider is the destination.
    IS_NOT_DEST = 0b1101    # run the CP if there is a propagation.
    IS_ENTRY = 0b1011       # run the CP if the provider is the entry point.
    IS_NOT_ENTRY = 0b0111   # run the CP if the provider is not entry point.

    # A recurring flag setting :
    IS_LINK = 0b0101        # the provider is nor the entry, nor the dest.

    propagation = None
    _sentinel = False

    methods = None

    @classmethod
    def _attach_to(cls, app):
        cls._router._app = app

    @classmethod
    def _propagate(cls):
        # recursive method called to registed all declared propagation rules.
        # this method is not intended to be used by the user.
        if cls.propagation is not None and not cls._sentinel:
            for endpoint, rule in cls.propagation.items():
                if not isinstance(rule, tuple):
                    rule = rule,

                provider, *urls = rule
                cls.add_propagation(provider, *urls, endpoint=endpoint)\
                   ._propagate()
            cls._sentinel = True    # recursion failsafe

    @classmethod
    def add_propagation(cls, provider, *urls, methods=None, endpoint=None):
        """Add a propagation rule.

        Args:
            provider: a provider class or an app-level endpoint.
            *urls: URL rules to access the target provider.
                If none is provided the propagation rule will fall back into
                app-level registered rules.
            methods: a tuple containing the allowed HTTP methods.
                If methods is none, the allowed methods are determined from
                the provider implementation.
            endpoint: the propagation endpoint.
                If the endpoint is none, the provider's name is used.

        Returns:
            The target provider.
        """
        app = cls._router._app

        if app is None:
            raise RuntimeError('%s is not reachable.' % cls.__name__)

        # convert string to provider
        target = None
        if not isinstance(provider, _ProviderType):
            target = provider
            provider = app.router.get_provider(provider)

        if provider is None:
            err_msg = '%s -> %s : unknown provider.'
            raise RuntimeError(err_msg % (cls.__name__, target))

        if target is None:
            target = provider.__name__

        if not urls:
            urls = tuple(app.router.iter_rules(target))

        if not urls:
            err_msg = '%s -> %s : explicit URLs required'
            raise RuntimeError(err_msg % (cls.__name__, provider.__name__))

        if endpoint is None:
            endpoint = provider.__name__
        # ensure the user is not messing with automatic endpoints
        endpoint = endpoint.replace('__', '_%s__' % cls.__name__)

        if methods is None:
            methods = provider.methods

        # register main URLs
        cls._router.add_provider(provider, *urls, methods=methods,
                                 endpoint=endpoint, propagation=True)

        return provider

    @classmethod
    def propagate(cls, *urls, methods=None, endpoint=None):
        """A class decorator used to declare a propagation rule.

        This does the same thing as add_propagation() but is intended for
        decorator usage. If no argument is submitted, the app-level URL rules
        of the decorated provider are used:

            @Provider0.propagate
            class Provider1(Provider):
                ...

            @Provider0.propagate('/custom-url')
            class Provider2(Provider):
                ...

        The @propagate decorator uses the same arguments as add_propagation().
        """
        def decorator(provider, _urls=None):
            if _urls is None:
                _urls = urls

            _endpoint = endpoint or provider.__name__

            if cls.propagation is None:
                cls.propagation = {}

            if endpoint in cls.propagation:
                err_msg = '%s : endpoint %s is already in use'
                raise RuntimeError(err_msg % (cls.__name__, _endpoint))

            cls.propagation[_endpoint] = (provider,) + _urls
            return provider

        if urls and isinstance(urls[0], _ProviderType):
            return decorator(urls[0], urls[1:])

        return decorator

    @asyncio.coroutine
    def _call_meth(self, match_info, name):
        # call meth with variable segments of the request as arguments.
        meth = getattr(self, name)
        if (not asyncio.iscoroutinefunction(meth) and
                not inspect.isgeneratorfunction(meth)):
            meth = asyncio.coroutine(meth)

        # get variable segments for the current provider.
        var = {k: v for k, v in match_info.items() if not k.startswith('_')}

        # get method signature and apply variable segments
        req, _, kw, _ = inspect.getargspec(getattr(meth, '__wrapped__', meth))
        if kw is None:
            rv = yield from meth(**{k: v for k, v in var.items() if k in req})
        else:
            rv = yield from meth(**var)  # any kerword arguments is accepted
        return rv

    @asyncio.coroutine
    def _execute_checkpoints(self, strategy, verb, match_info):
        # execute all the checkpoints of the selected strategy
        checkpoints = self._checkpoints.get(strategy, ())
        for name, verbs in checkpoints:
            if verbs is None or verb in verbs:
                yield from self._call_meth(match_info, name)

    @asyncio.coroutine
    def local_dispatcher(self, match_info, method, entry=True):
        """Dispatch the request for processing."""
        path = match_info.get('__pps')
        if path is not None:
            path = '/%s' % path     # add required leading /
            target = yield from self._router.resolve(None, path, method)

            # get the applied rule (remove __pps_ prefix if necessary)
            endpoint = target.endpoint
            if endpoint.startswith('__pps_'):
                endpoint = endpoint[6:]

            # select the checkpoint strategy from entry
            strategy = '__%s_%s' % (('lnk', 'src')[entry], endpoint)

            yield from self._execute_checkpoints(strategy, method, match_info)
            rv = yield from target._provider.local_dispatcher(target, method,
                                                              False)
            return rv

        if entry:
            yield from self._execute_checkpoints('__sngl', method, match_info)
        else:
            yield from self._execute_checkpoints('__dest', method, match_info)

        rv = yield from self._call_meth(match_info, method.lower())
        return rv



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Task context global variables."""


import tasklocals
from functools import partial


class LocalStorage:
    """A key-value data storage bind with a task context."""

    _app = None
    _task = None

    @classmethod
    def bind(cls, app):
        """Bind the storage to the app."""
        cls._app = app
        cls._task = tasklocals.local(loop=app.loop)

    def app(self):
        """Return the app."""
        return self._app

    def set(self, name, attr):
        """Set an attribute to the task context."""
        if self._task is None:  # pragma: no cover
            raise RuntimeError('LocalStorage is not bind to an application.')
        setattr(self._task, name, attr)

    def get(self, name, default=...):
        """Get an attribute from the task context."""
        if self._task is None:
            raise RuntimeError('LocalStorage is not bind to an application.')
        rv = getattr(self._task, name, default)

        if rv is Ellipsis:  # pragma: no cover
            raise KeyError(name)
        return rv

    def delete(self, name):  # pragma: no cover
        """Delete an attribute from the task context."""
        if self._task is None:
            raise RuntimeError('LocalStorage is not binded to an application')
        delattr(self._task, name)


class _Proxy:

    def __init__(self, proxy):
        self._proxy = proxy

    def __getattr__(self, attr):
        proxy = self._proxy()
        return getattr(proxy, attr)

    def __eq__(self, other):
        return other is self._proxy()

    def __repr__(self):
        proxy = self._proxy()
        return '<Proxy: %s>' % proxy.__repr__()


request = _Proxy(partial(LocalStorage().get, 'request'))
current_app = _Proxy(partial(LocalStorage().app))



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Implements all of the security mechanisms in use by the providers."""


import inspect


ALWAYS = 0b1111     # ALWAYS is the default scope and the initialization value


class ScopeError(Exception):
    """An exception for invalid scopes."""

    def __init__(self, message, function):
        """Initialize the exception."""
        _file = inspect.getfile(function)
        _, _line = inspect.getsourcelines(function)

        err_msg = '%s (File "%s", line %s)'
        super().__init__(err_msg % (message, _file, _line))


def checkpoint(scope=ALWAYS, priority=10, methods=None):
    """Decorate a provider method to make it a security checkpoint.

    A security checkpoint is a method of a provider executed each time the
    latter is required to process a request, to verify if prerequisites are
    met. If those prerequisites aren't met, the security checkpoint has the
    responsability to gracefully terminate the request with a call to abort().

    Each variable parts of an URL is passed as keyword argument to the
    decorated method. The unexpected keyword arguments are simply ignored.

    A security checkpoint possesses three attributes : scope, priority and
    methods. The scope is the combination of one or more values to determine
    when the security checkpoint should be executed. The valid values are a
    named propagation rule of the provider or the following flags:

    - Provider.ALWAYS:
      The security checkpoint is always executed. (default value)

    - Provider.IS_DEST:
      The security checkpoint is executed each time a propagation rule of the
      provider is executed.

    - Provider.IS_ENTRY:
      The security checkpoint is executed each time the provider is the one
      selected by the application as the entry-point for the processing.

    - Provider.IS_NOT_DEST:
      The negation of Provider.IS_DEST.

    - Provider.IS_NOT_ENTRY:
      The negation of Provider.IS_ENTRY.

    Each flag is cumulative with the help of '&' logical operator. Beware,
    some flags are incompatible and could raise an exception !

    The priority determines the execution order of each security checkpoint in
    the current scope. A lower number is an higher priority level. The default
    level is set to 10.

    The methods is used to limit the security checkpoint to certain HTTP verbs.
    It accepts the name of one verb or a tuple for more of them. The default
    value is None and is used to disable the limitation to certain HTTP verbs.

    Usages:

            @checkpoint
            def default_checkpoint(self, id):
                ...

            @checkpoint(scope=Provider.IS_ENTRY & Provider.IS_DEST)
            def single_scope_checkpoint(self, name=None):
                ...

            @checkpoint(scope=(Provider.IS_ENTRY, 'propagation_rule'))
            def multiple_scope_checkpoint(self, **kwargs):
                ...

            @checkpoint(priority=1, methods='GET')
            def high_priority_get_checkpoint(self):
                ...
    """
    if hasattr(scope, '__call__'):  # the decorator is used with default values
        func = scope
        setattr(func, '__is_checkpoint', ((ALWAYS,), priority, methods))
        return func

    if methods is not None and not isinstance(methods, tuple):
        methods = methods,

    if not isinstance(scope, tuple):
        scope = scope,

    def decorator(func):

        for desc in scope:
            if not isinstance(desc, int):
                continue

            if desc == ALWAYS and len(scope) > 1:
                raise ScopeError('ALWAYS is masking other scopes.', func)

            if not (desc % 4 and desc >> 2):
                raise ScopeError('Conflicting scope detected.', func)

        setattr(func, '__is_checkpoint', (scope, priority, methods))
        return func

    return decorator



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Implements a RESTful provider for CRUD operations."""


import string
import random
import asyncio

from hypr.globals import request
from hypr.web_exceptions import abort
from hypr.providers.base import Provider
from hypr.helpers.mini_dsl import parse_clause
from collections import defaultdict


class _Query:

    def __init__(self, filters, limit, offset, sort):
        if not isinstance(sort, tuple):
            sort = tuple(sort)

        self.filters = filters
        self.limit = limit
        self.offset = offset
        self.sort = sort


class CRUDProvider(Provider):
    """RESTful CRUD provider.

    The provider implements all the required methods to perform CRUD operations
    on the model associated with it through the HTTP protocol. It should
    perform nicely with most of the use-cases but non-trivial models may
    require additional development.

    The following features are provided out of the box :

    - searching, filtering and pagination on GET requests with transparent
      support for range filtering and dates.
    - bulk operations on POST, PUT/PATCH and DELETE.
    - atomicity of each request is guaranteed for transactional backends.

    Attributes:
        __model__: A Model associated with the current provider.
    """

    __model__ = None

    def url_to_uid(self, **kwargs):
        """Convert the URL arguments to a unique resource identifier.

        The unique identifier (uid) of a resource is a set of values unique to
        each instance of a model. The default implementation of url_to_uid()
        maps each URL arguments to an element of the model's key otherwise a
        custom implementation must be done.

        Returns: a tuple.

        Raises:
            NotImplementedError: custom implementation required.
        """
        key = self.__model__._key()
        rv = tuple(kwargs.get(cut) for cut in key if cut in kwargs)

        # check if the URL, mapping and uid match
        if len(kwargs) != len(key) or len(kwargs) != len(rv):
            err_msg = '%s : custom url_to_uid() is required'
            raise NotImplementedError(err_msg % self.__class__.__name__)

        return rv

    @staticmethod
    def parse_query(args):
        """Parse the query string of the request."""
        app = request.app

        # ensure offset is a positive integer
        offset = args.get('_offset', '0')
        if not offset.isdigit():
            raise ValueError

        # ensure limit is a positive integer
        limit = args.get('_limit', '0')
        if not limit.isdigit():
            raise ValueError

        limit = int(limit) or app.config.get('CRUD_DEFAULT_LIMIT', 0)
        absolute = app.config.get('CRUD_ABSOLUTE_LIMIT', limit)

        # the random prefix is used to avoid collision between user-defined
        # groups and automatic groups in HMQ
        random_prefix = ''.join(random.SystemRandom().choice(
            string.ascii_uppercase + string.digits) for _ in range(8))

        # get the filters
        filters = defaultdict(tuple)
        for group, (key, value) in enumerate(args.items()):
            if key.startswith('_'):
                continue
            group = '%s-%s' % (random_prefix, group)
            # use explicit group if one exists in the query
            value, group, *_ = value.rsplit(':', 1) + [group]
            filters[key] += parse_clause(value, group)

        # get the sorting keys
        sort = tuple(k for k in request.args.get('_sort', '').split(',') if k)
        if '-' in sort:
            raise ValueError

        return _Query(
            filters=dict(filters),
            limit=min(limit, absolute),
            offset=max(int(offset), 0),
            sort=sort
        )

    @asyncio.coroutine
    def get(self, **kwargs):
        """Get a specific resource or a collection of resources."""
        model = self.__model__

        # get a specific resource.
        if len(kwargs):
            return model.one(*self.url_to_uid(**kwargs)) or abort(404)

        try:
            query = self.parse_query(request.args)
        except ValueError:
            abort(400)

        # build the collection response object
        rv = {
            'count': model.count(**query.filters),
            'content': model.get(_limit=query.limit,
                                 _offset=query.offset,
                                 _sort=query.sort,
                                 **query.filters)
        }

        return rv

    @asyncio.coroutine
    def post(self, _data=None, _bulk=None, _commit=True):
        """Create a new resource."""
        model = self.__model__

        # get the _data from the request if None is given as method args.
        if _data is None:
            _data = yield from request.json()
        if _data is None:
            abort(400)

        # determine if this is a bulk insert or not.
        if _bulk is None:
            _bulk = request.args.get('_bulk', '').lower() in ['true', '1']

        # bulk insertion is based on the basic insertion
        if _bulk:
            content = []
            for resource in _data:
                rv, _ = yield from self.post(
                    _data=resource, _bulk=False, _commit=False)
                content.append(rv)

            rv = {
                'content': content,
                'count': len(content)
            }

        # create one resource
        # TODO: validation
        else:
            try:
                rv = model(**_data)
            except TypeError:
                model.rollback()
                abort(400)

            rv.save(commit=False)   # mark the resource for persistence

        # commit data
        # TODO: handle commit failure
        if _commit:
            model.commit()

        return rv, 201

    def patch(self, _data=None, _bulk=None, **kwargs):
        """The patch() method is an alias to put()."""
        return self.put(_data=_data, _bulk=_bulk, **kwargs)

    @asyncio.coroutine
    def put(self, _data=None, _bulk=None, _commit=True, **kwargs):
        """Update an existing resource."""
        model = self.__model__

        # get the _data from the request if None is given as method args.
        if _data is None:
            _data = yield from request.json()
        if _data is None:
            abort(400)

        # determine if this is a bulk insert or not.
        if _bulk is None:
            _bulk = request.args.get('_bulk', '').lower() in ['true', '1']

        if _bulk and not kwargs:
            content = []
            key = model._key()
            for resource in _data:
                url_args = {k: v for k, v in resource.items() if k in key}
                if len(url_args) != len(key):
                    model.rollback()
                    abort(400)
                rv = yield from self.put(
                    _data=resource, _bulk=False, _commit=False, **url_args)
                content.append(rv)

            rv = {
                'content': content,
                'count': len(content)
            }

        # update a resource
        elif not _bulk and kwargs:
            rv = model.one(*self.url_to_uid(**kwargs))  # retrieve the resource
            if rv is None and not _commit:
                model.rollback()
                abort(400)
            elif rv is None:
                model.rollback()
                abort(404)  # otherwise, 404 NOT FOUND

            # update the resource attributes
            for k, v in _data.items():
                if hasattr(rv, k):
                    if getattr(rv, k) != v:
                        setattr(rv, k, v)
                else:
                    model.rollback()
                    abort(400)

            rv.save(commit=False)

        # invalid request
        else:
            abort(400)

        # commit data
        # TODO: handle commit failure
        if _commit:
            model.commit()

        return rv

    @asyncio.coroutine
    def delete(self, _data=None, _bulk=None, _commit=True, **kwargs):
        """Delete an existing resource."""
        model = self.__model__

        # determine if this is a bulk insert or not.
        if _bulk is None:
            _bulk = request.args.get('_bulk', '').lower() in ['true', '1']

        # bulk delete
        if _bulk and not kwargs:

            # get the _data from the request if None is given as method args.
            if _data is None:
                _data = yield from request.json()
            if _data is None:
                abort(400)

            content = []
            key = model._key()
            for resource in _data:
                url_args = {k: v for k, v in resource.items() if k in key}
                if len(url_args) != len(key):
                    model.rollback()
                    abort(400)
                rv = yield from self.delete(
                    _bulk=False, _commit=False, **url_args)
                content.append(rv)

        # delete a resource
        elif not _bulk and kwargs:
            rv = model.one(*self.url_to_uid(**kwargs))  # retrieve the resource
            # if commit is not active, rollback all the transaction on failure
            if rv is None and not _commit:
                model.rollback()
                abort(400)
            elif rv is None:
                abort(404)  # otherwise, 404 NOT FOUND

            rv.delete(commit=False)

        # invalid request
        else:
            abort(400)

        # commit data
        # TODO: handle commit failure
        if _commit:
            model.commit()
            return '', 204

        return rv



#!/usr/bin/env python3
"""Build a secure and RESTful hypermedia API."""


import re
import sys
from pathlib import Path
from setuptools import setup, find_packages

if sys.version_info < (3, 4):
    raise RuntimeError("hypr requires Python 3.4+")

# read the version number from package
with (Path(__file__).resolve().parent / 'hypr/__init__.py').open() as f:
    v, = re.search(".*__version__ = '(.*)'.*", f.read(), re.MULTILINE).groups()


setup(

    name='hypr',
    version=v,

    author='Morgan Delahaye-Prat',
    author_email='mdp@sillog.net',
    maintainer='Morgan Delahaye-Prat',
    maintainer_email='mdp@sillog.net',

    url='https://project-hypr.github.io',
    description=__doc__,
    long_description=open('README.rst').read(),

    install_requires=open('requirements.txt').read().splitlines(),
    tests_require=open('requirements_test.txt').read().splitlines(),

    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    scripts=[],

    license='BSD',
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ]

)



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Provider related features."""


from aiohttp.web import Response
from hypr.providers.base import Provider
from hypr.providers.crud import CRUDProvider
from hypr.providers.security import checkpoint


__all__ = ['Provider', 'CRUDProvider', 'Response', 'checkpoint']



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Hypr Mini Query (HMQ) DSL

HMQ is a language to search and filter resources, fitted to the query string
of an URL. It supports unconstrained intervals, the logical negation and allows
to express complex logical expressions with the use of clausal normal form.


Predicates
----------

The HMQ provides the following predicates with x is a property of the
resources:

x=y         x is equal to y
x=!y        x is different from y
x=[y,z]     x is greater or equal to y and x is lower or equal to z
x=[y,]      x is greater or equal to y
x=[,z]      x is lower or equal to z
x=![y,z]    x is lower than y or x is greater than z
x=![y,]     x is lower than y
x=![,z]     x is greater than z

The following predicates are available when y and z are strings:

x=y*        x starts with y
x=*y        x ends with y
x=*y*       x contains y


Query construction
------------------

Each field-value pair in the query string is a clause with the field used as a
the property of the predicate.

URL: /path/to/resources?prop0=val0&prop1=val1
stands for: (resource.prop0 = val0) ∧ (resource.prop1 = val1)

It is possible to use multiple predicates on the same property in the same
clause with a logical OR:

URL: /path/to/resources?prop0=val0|val1
stands for: (resource.prop0 = val0 ∨ resource.prop0 = val1)

It is possible to group multiple predicates on different properties in the same
clause with the use of named group:

URL: /path/to/resources?prop0:group0=val0&prop1:group0=val1
stands for: (resource.prop0 = val0 ∨ resource.prop1 = val1)

Any query string complying to this scheme is a valid query:

?property[:group]=predicate[|predicate...][&clause...]


Python notation
---------------

Each query is writable in python with each predicate described in the form of
a tuple of three elements also known as the full notation:

(negation, predicate, group)

it also exists the partial notation `(negation, predicate)` and the simple
notation `(predicate,)` or `predicate`

Each predicates associated to a same property are grouped in a tuple:

resource.get(prop0=val0, prop1=val1)
stands for: (resource.prop0 = val0) ∧ (resource.prop1 = val1)

resource.get(prop0=(val0, val1))
resource.get(prop0=((True, val0), (True, val1)))
stands for: (resource.prop0 = val0 ∨ resource.prop0 = val1)

resources.get(prop0=(True, val0, 'group0'), prop1=(True, val1, 'group1'))
stands for: (resource.prop0 = val0 ∨ resource.prop1 = val1)

The normalized form is a tuple of full notations.


Known limitations
-----------------

A partial notation of two boolean values is not discernible from a tuple of two
simple boolean notations and will raise a RuntimeError as a non deterministic
query.
"""


import re
from datetime import datetime


DATE = re.compile('(\d{4}-\d{2}-\d{2})(?:T(\d{2}:\d{2}:\d{2})(?:.(\d+))?Z)?')
RANGE = re.compile('\[(.+)?(?(1),(.+)?|,(.+))\]')


class Range:
    """A custom range object with no type constaints."""

    def __init__(self, start=None, stop=None):
        """Initialization."""
        self.start = start
        self.stop = stop

    def __eq__(self, other):
        """Compare two range objects."""
        if not hasattr(other, 'start') or not hasattr(other, 'stop'):
            return False
        return self.start == other.start and self.stop == other.stop


def autocast(string):
    """Automatically cast a string to the best type."""
    if string is None:
        return None

    l_string = string.lower()

    if l_string == 'null':
        return None

    if l_string in ('true', 'on', 'yes'):
        return True

    if l_string in ('false', 'off', 'no'):
        return False

    try:
        numeric = float(string)
        integer = int(numeric)
        return (numeric, integer)[integer == numeric]
    except ValueError:
        pass

    re_match = DATE.match(string)
    if re_match is not None:
        val = re_match.groups()
        try:
            return datetime.strptime(
                '%sT%s.%sZ' % (val[0], val[1] or '00:00:00', val[2] or '0'),
                '%Y-%m-%dT%H:%M:%S.%fZ'
            )
        except ValueError:
            pass

    return string


def parse_predicate(string, group=None):
    """Parse a predicate string."""
    positive = string[0] != '!'
    value = string[not positive:]                    # >:-) pythonic ta mère !

    match = RANGE.match(value)
    if match is not None:
        start, stop = match.group(1), match.group(2) or match.group(3)
        value = Range(autocast(start), autocast(stop))
    elif '*' in value:
        value = re.compile('^%s$' % value.replace('*', '.*'))
    else:
        value = autocast(value)

    return positive, value, group


def parse_clause(string, group=None):
    """Parse a clause string."""
    return tuple(parse_predicate(p, group) for p in string.split('|') if p)


def normalize_query(query, key=None):
    """Normalize a query.

    A normalized query is expressed with a tuple of full notations.
    """
    if not isinstance(query, tuple):
        return ((True, query, key),)           # simple value

    query_length = len(query)

    if query_length == 0:                       # empty query
        return query

    if isinstance(query[0], bool):              # tuple start with a bool
        if 1 < query_length < 4 and isinstance(query[1], bool):
            raise RuntimeError('non determinist query.')
        elif 1 < query_length < 4:              # bool, other[, other]
            return ((query + (key,))[:3],)      # is a [partial] complexe value

    rv = []
    for value in query:
        if not isinstance(value, tuple):
            rv.append((True, value, key))
        else:
            rv.append((value + (key,))[:3])

    return tuple(rv)



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Hypr.

The security-oriented framework to build your micro-services.
"""


from hypr.app import Hypr
from hypr.providers import Provider, Response
from hypr.web_exceptions import abort


__version__ = '0.8.3'
__all__ = ('Hypr', 'Provider', 'Response', 'abort')



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""Request router."""


import re
import asyncio

from aiohttp.web import Response
from aiohttp.abc import AbstractRouter, AbstractMatchInfo
from hypr.globals import LocalStorage
from hypr.converters import DEFAULT_CONVERTERS
from hypr.serializers import choose_media_type
from hypr.web_exceptions import HTTPTemporaryRedirect, HTTPMethodNotAllowed, \
    HTTPNotFound, HTTPNotAcceptable


# regex to match the different segments of an URL rule
_rule_re = re.compile(r'''
    (?P<static>[^<]*)                           # static rule data
    <
    (?:
        (?P<converter>[a-zA-Z_][a-zA-Z0-9_]*)   # converter name
        (?:\((?P<args>.*?)\))?                  # converter arguments
        \:                                      # variable delimiter
    )?
    (?P<variable>[a-zA-Z_][a-zA-Z0-9_]*)        # variable name
    >
''', re.VERBOSE)

# regex to match converter arguments
_converter_args_re = re.compile(r'''
    ((?P<name>\w+)\s*=\s*)?
    (?P<value>
        True|False|
        -?(\d+.)?\d+(e[-+]\d+)?|
        -?\d+.|
        \w+|
        [urUR]?(?P<stringval>"[^"]*?"|'[^']*')|
    )\s*,
''', re.VERBOSE | re.UNICODE)

# translation table text -> python constants
_PYTHON_CONSTANTS = {
    'None':     None,
    'True':     True,
    'False':    False
}


def _pythonize(value):
    if value in _PYTHON_CONSTANTS:
        return _PYTHON_CONSTANTS[value]
    for convert in int, float:
        try:
            return convert(value)
        except ValueError:
            pass
    if value[:1] == value[-1:] and value[0] in '"\'':
        value = value[1:-1]
    return str(value)


def get_converter(converter_name, args, kwargs):
    """Look up the converter for the given parameter."""
    if converter_name not in DEFAULT_CONVERTERS:
        raise LookupError('The converter %s does not exist' % converter_name)
    return DEFAULT_CONVERTERS[converter_name](*args, **kwargs)


def parse_rule(rule):
    """Parse an URL rule.

    Parse a rule and return it as a generator. Each iteration yields tuples
    in the form ``(converter, arguments, variable)``. If the converter is
    `None` it's static url part, otherwise it's a dynamic one.
    """
    pos = 0
    end = len(rule)

    do_match = _rule_re.match
    used_names = set()

    while pos < end:
        match = do_match(rule, pos)
        if match is None:
            break

        # yield static segment
        data = match.groupdict()
        if data['static']:
            yield None, None, data['static']

        # yield variable segment
        variable = data['variable']
        converter = data['converter'] or 'default'
        if variable in used_names:
            raise ValueError('variable name %r used twice.' % variable)
        used_names.add(variable)
        yield converter, data['args'] or None, variable

        pos = match.end()

    if pos < end:
        remaining = rule[pos:]
        if '>' in remaining or '<' in remaining:
            raise ValueError('malformed url rule: %r' % rule)
        yield None, None, remaining


def parse_converter_args(argstr):
    """Parse the arguments of a converter."""
    argstr += ','
    args = []
    kwargs = {}

    for item in _converter_args_re.finditer(argstr):
        value = item.group('stringval')
        if value is None:
            value = item.group('value')
        value = _pythonize(value)
        if not item.group('name'):
            args.append(value)
        else:
            name = item.group('name')
            kwargs[name] = value

    return tuple(args), kwargs


class Rule:
    """The Rule object associates an URL rule to a endpoint.

    The URL rule is then used to match the submitted URL of incoming requests
    and determine the valid endpoint to handle them. An URL rule defines a set
    of static and parameterized segments.

    The rule object also stores additional informations, such as the accepted
    methods to accelerate checks.
    """

    def __init__(self, url, endpoint, methods=None):
        """Init the Rule.

        Args:
            url (str): an URL rule.
            endpoint: the endpoint associated to the rule.
            methods: tuple of the accepted methods. (None for any)
        """
        self.url = url
        self.endpoint = endpoint

        self.methods = None
        if methods is not None:
            self.methods = set([m.upper() for m in methods])

        self.arguments = set()
        self.is_leaf = not url.endswith('/')
        self._trace = self._converters = self._regex = self._weight = None

    def compile(self):
        """Compile the URL rule to a regular expression and store it."""
        self._trace = []
        self._weights = []
        self._converters = {}

        regex_parts = []

        def _build_regex(rule):

            for converter, arguments, variable in parse_rule(rule):
                if converter is None:
                    regex_parts.append(re.escape(variable))
                    self._trace.append((False, variable))
                    for part in variable.split('/'):
                        if part:
                            self._weights.append((0, -len(part)))
                else:
                    if arguments:
                        c_args, c_kwargs = parse_converter_args(arguments)
                    else:
                        c_args = ()
                        c_kwargs = {}
                    convobj = get_converter(converter, c_args, c_kwargs)
                    regex_parts.append('(?P<%s>%s)' % (variable,
                                                       convobj.regex))
                    self._converters[variable] = convobj
                    self._trace.append((True, variable))
                    self._weights.append((1, convobj.weight))
                    self.arguments.add(str(variable))

        _build_regex(self.is_leaf and self.url or self.url.rstrip('/'))
        if not self.is_leaf:
            self._trace.append((False, '/'))

        regex = r'^{}{}$'.format(
            ''.join(regex_parts),
            not self.is_leaf and '(?<!/)(?P<__suffix__>/?)' or ''
        )

        self._regex = re.compile(regex, re.UNICODE)

    def match(self, path):
        """Check if the rule matches a given path.

        If the rule matches a dict with the converted values is returned,
        otherwise the return value is `None`.
        """
        m = self._regex.search(path)
        if m is not None:
            groups = m.groupdict()

            if not self.is_leaf and not groups.pop('__suffix__'):
                raise HTTPTemporaryRedirect(path + '/')

            result = {}
            for name, value in groups.items():
                try:
                    value = self._converters[name].to_python(value)
                except ValueError:
                    return
                result[str(name)] = value

            return result

    def match_compare_key(self):
        """The match compare key for sorting.

        Current implementation:
        1.  rules without any arguments come first for performance
            reasons only as we expect them to match faster and some
            common ones usually don't have any arguments (index pages etc.)
        2.  The more complex rules come first so the second argument is the
            negative length of the number of weights.
        3.  lastly we order by the actual weights.
        """
        return bool(self.arguments), -len(self._weights), self._weights


class MatchInfo(dict, AbstractMatchInfo):
    """Matching informations of a Rule."""

    def __init__(self, match_dict, rule, provider, serializer, mimetype):
        """Init MatchInfo."""
        super().__init__(match_dict)
        self._provider = provider
        self._rule = rule
        self._serializer = serializer
        self._mimetype = mimetype
        self._apps = []
        self._frozen = False

    @property
    def apps(self):
        return tuple(self._apps)

    def add_app(self, app):
        if self._frozen:
            raise RuntimeError("Cannot change apps stack after .freeze() call")
        self._apps.insert(0, app)

    @property
    def endpoint(self):
        """Return the endpoint."""
        return self._rule.endpoint

    @property
    def handler(self):
        """Request handler."""
        def ensure_response(request):

            if self._serializer is None:
                raise HTTPNotAcceptable()

            rv = yield from self._provider.local_dispatcher(self, request.method)

            status_or_headers = headers = None
            if isinstance(rv, tuple):
                rv, status_or_headers, headers = rv + (None,) * (3 - len(rv))

            if isinstance(status_or_headers, (list, dict)):
                headers, status_or_headers = status_or_headers, None

            if not isinstance(rv, Response):

                data = rv
                rv = Response(headers=headers, status=status_or_headers or 200)
                rv.headers['content-type'] = self._mimetype     # set mimetype
                rv.text = self._serializer(data, request)
                headers = status_or_headers = None

            return rv

        return ensure_response

    def __repr__(self):     # pragma: no cover
        """Representation of the MatchInfo object."""
        return "<MatchInfo {}: {}>".format(super().__repr__(), self._rule)


class Router(AbstractRouter):
    """Hypr router."""

    def __init__(self, app):
        super().__init__()
        self._app = app
        self._rules = []
        self._providers = {}
        self._serializers = {}

    def _propagate(self):
        for provider in self._providers.values():
            provider.__class__._propagate()

    @asyncio.coroutine
    def resolve(self, request, path=None, method=None):
        """Return the match_info for the given request."""
        # determine the best mimetype/serializer for the response
        mimetype = None
        serializer = None

        if request is not None:
            path = request.path
            method = request.method
            app = self._app

            accept = request.headers.get('accept', '*/*')
            mimetype = choose_media_type(accept, app.available_mimetypes)
            serializer = self._serializers.get(mimetype)

            LocalStorage().set('request', request)

        allowed_methods = set()

        for rule in self._rules:
            # search the matching rule
            match_dict = rule.match(path)
            if match_dict is None:
                continue

            # get the provider if the rule handles the HTTP method
            if rule.methods is None or method in rule.methods:
                provider = self._providers[rule.endpoint]
                return MatchInfo(match_dict, rule, provider,
                                 serializer, mimetype)

            allowed_methods.update(rule.methods)

        if allowed_methods:
            raise HTTPMethodNotAllowed(method, allowed_methods)
        else:
            raise HTTPNotFound()

    def set_serializers(self, serializers):
        """Set the available serializers."""
        self._serializers = serializers

    def add_provider(self, provider, *urls, methods=None, endpoint=None,
                     propagation=False):
        """Register a provider with URL rules.

        Args:
            provider: a Provider class.
            *urls: one or more path to access the provider.
            methods: a tuple containing the allowed HTTP methods.
                If methods is none, the allowed methods are determined from
                the provider implementation.
            endpoint: the name of the endpoint. If endpoint is none, the name
                is the provider's name.
        """
        provider._attach_to(self._app)

        if endpoint is None:
            endpoint = provider.__name__    # default endpoint

        if methods is None:
            methods = provider.methods

        if endpoint in self._providers:
            err_msg = 'endpoint %s is already in use'
            raise RuntimeError(err_msg % endpoint)

        path = None
        for path in urls:

            if not path.startswith('/'):
                err_msg = 'url \'%s\' should start with a leading \'/\''
                raise RuntimeError(err_msg % path)

            rule = Rule(path, endpoint, methods)
            rule.compile()
            self._rules.append(rule)
            self._rules.sort(key=lambda x: x.match_compare_key())

        if path:
            self._providers[endpoint] = provider()

        # register propagation
        if propagation:
            endpoint = '__pps_%s' % endpoint
            urls = tuple(
                ('%s' % ('/', '')[url[-1] == '/']).join((url, '<path:__pps>'))
                for url in urls if '<path:' not in url
            )

            if urls:
                self.add_provider(provider, *urls, endpoint=endpoint)

    def get_provider(self, endpoint):

        rv = self._providers.get(endpoint)
        if rv is not None:
            return rv.__class__
        return None

    def iter_rules(self, endpoint=None):

        return (rule.url for rule in self._rules if endpoint is None or
                rule.endpoint == endpoint)



# Copyright 2014-2016 Morgan Delahaye-Prat. All Rights Reserved.
#
# Licensed under the Simplified BSD License (the "License");
# you may not use this file except in compliance with the License.
"""A parser for the Accept header."""


import re


accept_range = re.compile(r'''
    (?P<type>\S+)/                                    # mandatory type
    (?:(?P<vendor>[^\s;+]+)\.)?                       # optional vendor
    (?P<subtype>[^\s;+]+(?:[^\s;]+)?)                 # mandatory subtype
    (?:;(?:(?:[ ]*q=(?P<quality>\d(?:\.\d)?))|(?:[^s;]+=[^s;]+)))*  # quality
''', re.VERBOSE)


def parse_accept_header(accept):
    """Parse the accept header.

    Each accepted media type is yielded as a tuple with the type, subtype and
    quality. If the quality is not set, it is defaulted to 1.
    """
    for media_range in accept.split(','):
        match = accept_range.match(media_range.strip())
        if match is None:
            break

        data = match.groupdict()
        yield (data['type'], data['subtype'], float(data['quality'] or 1.0))


def choose_media_type(accept, available):
    """Choose the best available media type to satisfy the accept header.

    Args:
        accept (str): an HTTP Accept header (RFC 2616).
        available (list): a list of valid media types (RFC 6838).
    Returns:
        the most well suited media type.
    """
    rv = []
    for type_, subtype, quality in parse_accept_header(accept):
        media_type = '%s/%s' % (type_, subtype)
        if media_type in available:
            if quality == 1:
                return media_type   # a best solution has been found
            else:
                rv.append((quality, media_type))

        # the wildcards are used to set a default quality to the remaining
        # media types available. if a wildcard has the same quality of an
        # accepted media type, the quality of the former is lowered to promote
        # the latter. For a given quality the order is as follow :
        # explicit media type > wildcard on the subtype > wilcard on the type
        elif '*' in media_type:
            for media_type in available:
                if type_ == '*' and media_type not in rv:
                    rv.append((quality - 0.02, media_type))
                elif media_type.startswith(type_) and media_type not in rv:
                    rv.append((quality - 0.01, media_type))

    if rv:
        rv.sort(key=lambda x: -x[0])
        return rv[0][1]
